package ru.dbotthepony.mc.otm.client.render

import com.google.common.collect.ImmutableList
import com.mojang.blaze3d.pipeline.MainTarget
import com.mojang.blaze3d.platform.GlConst
import com.mojang.blaze3d.platform.GlStateManager
import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.BufferBuilder
import com.mojang.blaze3d.vertex.BufferUploader
import com.mojang.blaze3d.vertex.DefaultVertexFormat
import com.mojang.blaze3d.vertex.Tesselator
import com.mojang.blaze3d.vertex.VertexBuffer
import com.mojang.blaze3d.vertex.VertexFormat
import com.mojang.math.Matrix4f
import it.unimi.dsi.fastutil.ints.Int2ObjectArrayMap
import it.unimi.dsi.fastutil.ints.Int2ObjectFunction
import net.minecraft.client.Minecraft
import net.minecraft.client.renderer.FogRenderer
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.core.Vec3i
import net.minecraft.world.level.levelgen.XoroshiroRandomSource
import net.minecraft.world.level.material.FogType
import org.lwjgl.opengl.GL14
import ru.dbotthepony.mc.otm.ClientConfig
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.core.RGBAColor
import ru.dbotthepony.mc.otm.core.component1
import ru.dbotthepony.mc.otm.core.component2
import ru.dbotthepony.mc.otm.core.component3
import ru.dbotthepony.mc.otm.core.linearInterpolation
import ru.dbotthepony.mc.otm.milliTime
import java.lang.ref.WeakReference
import java.util.stream.Collectors
import kotlin.math.absoluteValue
import kotlin.math.ceil
import kotlin.math.pow

@Suppress("SameParameterValue")
object GlitchRenderer {
	private abstract class VideoGlitchType {
		abstract fun upload(builder: BufferBuilder, x: Double, y: Double, width: Double, height: Double, u0: Float, v0: Float, u1: Float, v1: Float)

		protected fun uploadVertices(faces: Int, builder: BufferBuilder, x: Double, y: Double, width: Double, height: Double, u0: Float, v0: Float, u1: Float, v1: Float, red: Int, green: Int, blue: Int, alpha: Int) {
			if (faces and BOTTOM_LEFT != 0) {
				builder.vertex(x, y + height, 0.0).uv(u0, v1).color(red, green, blue, alpha).endVertex()
			}

			if (faces and BOTTOM_RIGHT != 0) {
				builder.vertex(x + width, y + height, 0.0).uv(u1, v1).color(red, green, blue, alpha).endVertex()
			}

			if (faces and TOP_RIGHT != 0) {
				builder.vertex(x + width, y, 0.0).uv(u1, v0).color(red, green, blue, alpha).endVertex()
			}

			if (faces and TOP_LEFT != 0) {
				builder.vertex(x, y, 0.0).uv(u0, v0).color(red, green, blue, alpha).endVertex()
			}
		}

		protected fun uploadQuad(builder: BufferBuilder, x: Double, y: Double, width: Double, height: Double, u0: Float, v0: Float, u1: Float, v1: Float, red: Int, green: Int, blue: Int, alpha: Int) {
			uploadVertices(TOP_LEFT or TOP_RIGHT or BOTTOM_RIGHT, builder, x, y, width, height, u0, v0, u1, v1, red, green, blue, alpha)
			uploadVertices(TOP_LEFT or BOTTOM_RIGHT or BOTTOM_LEFT, builder, x, y, width, height, u0, v0, u1, v1, red, green, blue, alpha)
		}

		companion object {
			const val TOP_LEFT = 1
			const val TOP_RIGHT = 2
			const val BOTTOM_LEFT = 4
			const val BOTTOM_RIGHT = 8

			val vertices = intArrayOf(
				TOP_LEFT,
				TOP_RIGHT,
				BOTTOM_LEFT,
				BOTTOM_RIGHT,
			)
		}
	}

	private val glitchesSorted: List<List<VideoGlitchType>>
	private val glitchesTrisSorted: List<List<VideoGlitchType>>
	private val glitchesQuadsSorted: List<List<VideoGlitchType>>

	private fun selectGlitch(): List<VideoGlitchType> {
		for (list in glitchesSorted) {
			if (random.nextFloat() < 0.35f) {
				return list
			}
		}

		return glitchesSorted[glitchesSorted.size - 1]
	}

	private fun selectGlitchQuad(): List<VideoGlitchType> {
		for (list in glitchesQuadsSorted) {
			if (random.nextFloat() < 0.35f) {
				return list
			}
		}

		return glitchesQuadsSorted[glitchesQuadsSorted.size - 1]
	}

	private fun selectGlitchTriangle(): List<VideoGlitchType> {
		for (list in glitchesTrisSorted) {
			if (random.nextFloat() < 0.35f) {
				return list
			}
		}

		return glitchesTrisSorted[glitchesTrisSorted.size - 1]
	}

	init {
		val glitchesMapQuads = Int2ObjectArrayMap<ArrayList<VideoGlitchType>>()
		val glitchesMapTris = Int2ObjectArrayMap<ArrayList<VideoGlitchType>>()
		val glitchesMap = Int2ObjectArrayMap<ArrayList<VideoGlitchType>>()

		val colorVariants = ArrayList<Vec3i>()

		val flippersFloppers = listOf(
			{ u0: Float, v0: Float, u1: Float, v1: Float -> UVCoords(u0, v0, u1, v1) },
			{ u0: Float, v0: Float, u1: Float, v1: Float -> UVCoords(1f - u0, v0, 1f - u1, v1) },
			{ u0: Float, v0: Float, u1: Float, v1: Float -> UVCoords(u0, 1f - v0, u1, 1f - v1) },
			{ u0: Float, v0: Float, u1: Float, v1: Float -> UVCoords(1f - u0, 1f - v0, 1f - u1, 1f - v1) },
		)

		val colors = intArrayOf(0, 10, 30, 50, 80, 100, 120, 150, 180, 220, 230, 235, 245, 249, 255)

		for (red in colors)
			for (green in colors)
				for (blue in colors)
					if (red != 255 || green != 255 || blue != 255)
						colorVariants.add(Vec3i(red, green, blue))

		for ((r, g, b) in colorVariants) {
			for (winding in UVWindingOrder.distinct) {
				for ((i, flipflop) in flippersFloppers.withIndex()) {
					object : VideoGlitchType() {
						@Suppress("name_shadowing")
						override fun upload(builder: BufferBuilder, x: Double, y: Double, width: Double, height: Double, u0: Float, v0: Float, u1: Float, v1: Float) {
							val (u0_, v0_, u1_, v1_) = winding.translate(u0, v0, u1, v1)
							val (u0, v0, u1, v1) = flipflop.invoke(u0_, v0_, u1_, v1_)
							uploadQuad(builder, x, y, width, height, u0, v0, u1, v1, r, g, b, 255)
						}
					}.also {
						if (i == 0) {
							glitchesMapQuads
								.computeIfAbsent(255 - r + 255 - g + 255 - b, Int2ObjectFunction { ArrayList() })
								.add(it)
						}

						glitchesMap
							.computeIfAbsent(255 - r + 255 - g + 255 - b, Int2ObjectFunction { ArrayList() })
							.add(it)
					}
				}
			}
		}

		for (va in VideoGlitchType.vertices) {
			for (vb in VideoGlitchType.vertices) {
				if (va == vb) continue

				for (cb in VideoGlitchType.vertices) {
					if (cb == va || cb == vb) continue

					for ((r, g, b) in colorVariants) {
						for (winding in UVWindingOrder.distinct) {
							for ((i, flipflop) in flippersFloppers.withIndex()) {
								@Suppress("name_shadowing")
								object : VideoGlitchType() {
									val vertices = va or vb or cb

									override fun upload(builder: BufferBuilder, x: Double, y: Double, width: Double, height: Double, u0: Float, v0: Float, u1: Float, v1: Float) {
										val (u0_, v0_, u1_, v1_) = winding.translate(u0, v0, u1, v1)
										val (u0, v0, u1, v1) = flipflop.invoke(u0_, v0_, u1_, v1_)
										uploadVertices(vertices, builder, x, y, width, height, u0, v0, u1, v1, r, g, b, 255)
									}
								}.also {
									if (i == 0) {
										glitchesMapTris
											.computeIfAbsent(255 - r + 255 - g + 255 - b, Int2ObjectFunction { ArrayList() })
											.add(it)
									}

									glitchesMap
										.computeIfAbsent(255 - r + 255 - g + 255 - b, Int2ObjectFunction { ArrayList() })
										.add(it)
								}
							}
						}
					}
				}
			}
		}

		val toSort = glitchesMapQuads.entries.stream().collect(Collectors.toList())
		toSort.sortBy { it.key }
		glitchesQuadsSorted = toSort.stream().map { ImmutableList.copyOf(it.value) }.collect(ImmutableList.toImmutableList())

		val toSort2 = glitchesMapTris.entries.stream().collect(Collectors.toList())
		toSort2.sortBy { it.key }
		glitchesTrisSorted = toSort2.stream().map { ImmutableList.copyOf(it.value) }.collect(ImmutableList.toImmutableList())

		val toSort3 = glitchesMap.entries.stream().collect(Collectors.toList())
		toSort3.sortBy { it.key }
		glitchesSorted = toSort3.stream().map { ImmutableList.copyOf(it.value) }.collect(ImmutableList.toImmutableList())
	}

	private val random = XoroshiroRandomSource(System.nanoTime(), System.currentTimeMillis())

	var redShiftX = 0.0
		private set
	var redShiftY = 0.0
		private set
	var greenShiftX = 0.0
		private set
	var greenShiftY = 0.0
		private set
	var blueShiftX = 0.0
		private set
	var blueShiftY = 0.0
		private set

	var lastGlitch = System.nanoTime()
		private set

	var nextGlitch = 0L
		private set

	var lastEncodingGlitch = System.nanoTime()
		private set

	var nextEncodingGlitch = 0L
		private set



	private val glitchBuffer by lazy(LazyThreadSafetyMode.NONE) {
		MainTarget(minecraft.window.width, minecraft.window.height)
	}

	private fun upload(builder: BufferBuilder, offsetX: Double, offsetY: Double, u0: Float, v0: Float, u1: Float, v1: Float) {
		builder.vertex(offsetX - 1.0, offsetY + 1.0, 0.0).uv(u0, v1).endVertex()
		builder.vertex(offsetX + 1.0, offsetY + 1.0, 0.0).uv(u1, v1).endVertex()
		builder.vertex(offsetX + 1.0, offsetY - 1.0, 0.0).uv(u1, v0).endVertex()
		builder.vertex(offsetX - 1.0, offsetY - 1.0, 0.0).uv(u0, v0).endVertex()
	}

	private fun uploadEmpty(builder: BufferBuilder, offsetX: Double, offsetY: Double) {
		builder.vertex(offsetX - 1.0, offsetY + 1.0, 0.0).endVertex()
		builder.vertex(offsetX + 1.0, offsetY + 1.0, 0.0).endVertex()
		builder.vertex(offsetX + 1.0, offsetY - 1.0, 0.0).endVertex()
		builder.vertex(offsetX - 1.0, offsetY - 1.0, 0.0).endVertex()
	}

	private fun uploadLine(builder: BufferBuilder, offsetX: Double, offsetY: Double, height: Double, u0: Float, v0: Float, u1: Float, v1: Float) {
		builder.vertex(offsetX - 1.0, offsetY - height, 0.0).uv(u0, v1).endVertex()
		builder.vertex(offsetX + 1.0, offsetY - height, 0.0).uv(u1, v1).endVertex()
		builder.vertex(offsetX + 1.0, offsetY, 0.0).uv(u1, v0).endVertex()
		builder.vertex(offsetX - 1.0, offsetY, 0.0).uv(u0, v0).endVertex()
	}

	private inline fun makeMirrors(handler: (x: Double, y: Double, u0: Float, v0: Float, u1: Float, v1: Float) -> Unit) {
		handler(0.0 - 2f, 0.0, 1f, 0f, 0f, 1f)
		handler(0.0 + 2f, 0.0, 1f, 0f, 0f, 1f)

		handler(0.0, 2.0, 0f, 1f, 1f, 0f)
		handler(0.0, -2.0, 0f, 1f, 1f, 0f)

		handler(0.0 - 2f, -2.0, 1f, 1f, 0f, 0f)
		handler(0.0 + 2f, -2.0, 1f, 1f, 0f, 0f)

		handler(0.0 - 2f, +2.0, 1f, 1f, 0f, 0f)
		handler(0.0 + 2f, +2.0, 1f, 1f, 0f, 0f)
	}

	private fun draw(offsetX: Double, offsetY: Double) {
		val builder = tesselator.builder

		builder.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION_TEX)

		upload(builder, offsetX, offsetY, 0f, 0f, 1f, 1f)

		if (offsetX != 0.0 || offsetY != 0.0) {
			makeMirrors { x: Double, y: Double, u0: Float, v0: Float, u1: Float, v1: Float ->
				upload(builder, offsetX + x, offsetY + y, u0, v0, u1, v1)
			}
		}

		BufferUploader.drawWithShader(builder.end())
	}

	private fun pixel2ViewX(value: Double) = linearInterpolation(value / glitchBuffer.width, 1.0, -1.0)
	private fun pixel2ViewY(value: Double) = linearInterpolation(value / glitchBuffer.height, 1.0, -1.0)
	private fun pixel2TextureX(value: Double) = (value / glitchBuffer.width).toFloat()
	private fun pixel2TextureY(value: Double) = 1f - (value / glitchBuffer.height).toFloat()

	private fun drawVHSLineGap(y: Double, height: Double) {
		val builder = tesselator.builder

		val v = pixel2TextureY(y)

		builder.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION_TEX)
		uploadLine(builder, 0.0, pixel2ViewY(y), (height / glitchBuffer.height) * 2.0, 0f, v, 1f, v)
		BufferUploader.drawWithShader(builder.end())
	}

	private val colorGlitchVertexBuffer by lazy {
		VertexBuffer()
	}

	private var colorGlitchBuff = arrayOfNulls<VideoGlitchType>(0)
	private var colorGlitchWidth = 0
	private var colorGlitchHeight = 0

	private fun putMacroGlitch(x: Int, y: Int, width: Int, height: Int, value: VideoGlitchType?) {
		for (offsetX in 0 until width) {
			for (offsetY in 0 until height) {
				val index = (x + offsetX) + (y + offsetY) * colorGlitchWidth

				if (colorGlitchBuff.size > index) {
					colorGlitchBuff[index] = value
				}
			}
		}
	}

	private fun colorGlitchChanceMultiplier(x: Int, y: Int): Float {
		return (1f - ((x.toDouble() - colorGlitchWidth.toDouble() / 2.0).pow(2.0) / colorGlitchWidth.toDouble().pow(2.0) +
				(y.toDouble() - colorGlitchHeight.toDouble() / 2.0).pow(2.0) / colorGlitchHeight.toDouble().pow(2.0))
				.toFloat()
				.coerceAtLeast(0f)
				.coerceAtMost(1f))
				.pow(1f / 8f)
	}

	private fun makeColorGlitch() {
		if (!ClientConfig.ENABLE_COLOR_NOISE) {
			return
		}

		lastEncodingGlitch = System.nanoTime()
		nextEncodingGlitch = random.nextIntBetweenInclusive(10_000_000, 40_000_000).toLong()
		colorGlitchWidth = ceil(glitchBuffer.width / GLITCH_BLOCK_SIZE).toInt()
		colorGlitchHeight = ceil(glitchBuffer.height / GLITCH_BLOCK_SIZE).toInt()

		colorGlitchBuff = arrayOfNulls((colorGlitchWidth + 1) * (colorGlitchHeight + 1))

		for (xBlock in 0 .. colorGlitchWidth)
			for (yBlock in 0 .. colorGlitchHeight)
				if (colorGlitchChanceMultiplier(xBlock, yBlock) < 0.97f && random.nextFloat() > colorGlitchChanceMultiplier(xBlock, yBlock) * 0.999f)
					colorGlitchBuff[xBlock + yBlock * colorGlitchWidth] = selectGlitch().let { it[random.nextInt(0, it.size)] }

		for (xBlock in 0 .. colorGlitchWidth)
			for (yBlock in 0 .. colorGlitchHeight)
				if (colorGlitchChanceMultiplier(xBlock, yBlock) < 0.97f && random.nextFloat() > colorGlitchChanceMultiplier(xBlock, yBlock) * 0.9999f)
					putMacroGlitch(xBlock, yBlock, random.nextInt(1, 5), random.nextInt(1, 5), selectGlitchQuad().let { it[random.nextInt(0, it.size)] })

		for (xBlock in 0 .. colorGlitchWidth)
			for (yBlock in 0 .. colorGlitchHeight)
				if (colorGlitchChanceMultiplier(xBlock, yBlock) < 0.97f && random.nextFloat() > colorGlitchChanceMultiplier(xBlock, yBlock) * 0.99999f)
					putMacroGlitch(xBlock, yBlock, random.nextInt(1, 9), random.nextInt(1, 9), selectGlitchQuad().let { it[random.nextInt(0, it.size)] })

		for (xBlock in 0 .. colorGlitchWidth)
			for (yBlock in 0 .. colorGlitchHeight)
				if (colorGlitchChanceMultiplier(xBlock, yBlock) < 0.97f && random.nextFloat() > colorGlitchChanceMultiplier(xBlock, yBlock) * 0.999999f)
					putMacroGlitch(xBlock, yBlock, random.nextInt(1, 14), random.nextInt(1, 14), selectGlitchQuad().let { it[random.nextInt(0, it.size)] })

		val blockStepWidth = (GLITCH_BLOCK_SIZE / glitchBuffer.width) * 2.0
		val blockStepHeight = (GLITCH_BLOCK_SIZE / glitchBuffer.height) * 2.0

		val blockStepWidthUV = GLITCH_BLOCK_SIZE / glitchBuffer.width
		val blockStepHeightUV = GLITCH_BLOCK_SIZE / glitchBuffer.height

		val builder = tesselator.builder

		builder.begin(VertexFormat.Mode.TRIANGLES, DefaultVertexFormat.POSITION_TEX_COLOR)

		for (xBlock in 0 .. colorGlitchWidth) {
			for (yBlock in 0 .. colorGlitchHeight) {
				colorGlitchBuff[xBlock + yBlock * colorGlitchWidth]?.upload(
					builder = builder,
					x = -1.0 + xBlock * blockStepWidth,
					y = -1.0 + yBlock * blockStepHeight,
					width = blockStepWidth,
					height = blockStepHeight,
					u0 = xBlock * blockStepWidthUV,
					v0 = yBlock * blockStepHeightUV,
					u1 = (xBlock + 1) * blockStepWidthUV,
					v1 = (yBlock + 1) * blockStepHeightUV,
				)
			}
		}

		val colorGlitchBuffer = builder.end()
		colorGlitchVertexBuffer.bind()
		colorGlitchVertexBuffer.upload(colorGlitchBuffer)
	}

	private fun makeGlitch() {
		redShiftX = random.nextDouble() * 0.05 - 0.025
		redShiftY = random.nextDouble() * 0.05 - 0.025
		greenShiftX = random.nextDouble() * 0.05 - 0.025
		greenShiftY = random.nextDouble() * 0.05 - 0.025
		blueShiftX = random.nextDouble() * 0.05 - 0.025
		blueShiftY = random.nextDouble() * 0.05 - 0.025
		lastGlitch = System.nanoTime()
		nextGlitch = random.nextIntBetweenInclusive(75_000_000, 400_000_000).toLong()
	}

	var glitchUntil = 0L
		private set
	var glitchSince = System.nanoTime()
		private set

	fun glitchFor(millis: Long) {
		if (glitchUntil - (System.nanoTime() - glitchSince) < millis) {
			glitchSince = System.nanoTime()
			glitchUntil = millis * 1_000_000L
		}
	}

	private const val GLITCH_BLOCK_SIZE = 8f

	private var lastAndroid: WeakReference<MatteryPlayerCapability>? = null

	@JvmStatic
	fun render() {
		if (System.nanoTime() - glitchSince >= glitchUntil && ((minecraft.player?.matteryPlayer ?: lastAndroid?.get())?.isAndroid != true || minecraft.player?.isAlive != false)) {
			return
		}

		if (lastAndroid?.get() !== minecraft.player?.matteryPlayer && minecraft.player?.matteryPlayer != null)
			lastAndroid = WeakReference(minecraft.player?.matteryPlayer ?: throw ConcurrentModificationException())

		val glitchBuffer = glitchBuffer
		val projection = RenderSystem.getProjectionMatrix()
		RenderSystem.setProjectionMatrix(Matrix4f().also { it.setIdentity() })

		RenderSystem.getModelViewStack().also {
			it.pushPose()
			it.setIdentity()
		}

		RenderSystem.applyModelViewMatrix()

		RenderSystem.disableCull()
		RenderSystem.disableDepthTest()
		RenderSystem.enableBlend()
		RenderSystem.enableTexture()

		if (glitchBuffer.width != minecraft.window.width || glitchBuffer.height != minecraft.window.height) {
			glitchBuffer.resize(minecraft.window.width, minecraft.window.height, Minecraft.ON_OSX)
		}

		if (System.nanoTime() - lastGlitch >= nextGlitch) {
			makeGlitch()
		}

		if (ClientConfig.ENABLE_COLOR_NOISE && System.nanoTime() - lastEncodingGlitch >= nextEncodingGlitch) {
			makeColorGlitch()
		}

		if (!Minecraft.useShaderTransparency() && minecraft.gameRenderer.mainCamera.fluidInCamera == FogType.NONE) {
			glitchBuffer.bindWrite(true)
			RenderSystem.clearColor(FogRenderer.fogRed, FogRenderer.fogGreen, FogRenderer.fogBlue, 1f)
			RenderSystem.clear(GlConst.GL_COLOR_BUFFER_BIT, Minecraft.ON_OSX)

			// fix main rendertarget transparency (which is everything on screen that is not covered by anything, unless running Fabulous! graphics option)
			// this is required because Minecraft's builtin fragement shaders are stupid
			// and discard fully transparent fragments
			// when, in fact, alpha channel is just another color channel, and it's meaning is dictated
			// by render context, and it is not strictly "transparency"
			// the downside of this is RGBA8888 texture double copy, requiring a lot of video memory bandwidth
			// and no, trying to render "fog" solid color overlay quad over main rendertarget with proper blend mode
			RenderSystem.setShaderTexture(0, minecraft.mainRenderTarget.colorTextureId)
			RenderSystem.blendFuncSeparate(GlStateManager.SourceFactor.SRC_ALPHA, GlStateManager.DestFactor.ONE_MINUS_SRC_ALPHA, GlStateManager.SourceFactor.ZERO, GlStateManager.DestFactor.ONE)
			RenderSystem.setShader(GameRenderer::getPositionTexShader)

			RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
			draw(0.0, 0.0)
			minecraft.mainRenderTarget.bindWrite(true)

			RenderSystem.setShader(GameRenderer::getPositionTexShader)

			RenderSystem.blendFunc(GlStateManager.SourceFactor.ONE, GlStateManager.DestFactor.ZERO)
			RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
			RenderSystem.setShaderTexture(0, glitchBuffer.colorTextureId)
			draw(0.0, 0.0)
		}

		glitchBuffer.bindWrite(true)
		RenderSystem.clearColor(0f, 0f, 0f, 1f)
		RenderSystem.clear(GlConst.GL_COLOR_BUFFER_BIT, Minecraft.ON_OSX)

		// distort colors by sampling main frame buffer (raw stage)
		RenderSystem.setShaderTexture(0, minecraft.mainRenderTarget.colorTextureId)
		RenderSystem.blendFuncSeparate(GlStateManager.SourceFactor.ONE, GlStateManager.DestFactor.ONE, GlStateManager.SourceFactor.ZERO, GlStateManager.DestFactor.ONE)

		RenderSystem.setShader(GameRenderer::getPositionTexShader)

		RenderSystem.setShaderColor(1f, 0.1f, 0f, 1f)
		draw(redShiftX, redShiftY)

		RenderSystem.setShaderColor(0f, 1f, 0f, 1f)
		draw(greenShiftX, greenShiftY)

		RenderSystem.setShaderColor(0f, 0f, 1f, 1f)
		draw(blueShiftX, blueShiftY)

		minecraft.mainRenderTarget.bindWrite(true)

		// draw back to main frame buffer (prepare for post process)
		RenderSystem.blendFunc(GlStateManager.SourceFactor.ONE, GlStateManager.DestFactor.ZERO)
		RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
		RenderSystem.setShaderTexture(0, glitchBuffer.colorTextureId)
		draw(0.0, 0.0)

		// return to our frame buffer (post process stage)
		glitchBuffer.bindWrite(true)
		RenderSystem.setShaderTexture(0, minecraft.mainRenderTarget.colorTextureId)

		// color perception errors (eye-camera glitch)
		drawVHSLineGap((milliTime % glitchBuffer.height).toDouble(), glitchBuffer.height * 0.025)
		drawVHSLineGap(((milliTime + glitchBuffer.height / 2) % glitchBuffer.height).toDouble(), glitchBuffer.height * 0.075)
		drawVHSLineGap(((milliTime + glitchBuffer.height / 3) % glitchBuffer.height).toDouble(), glitchBuffer.height * 0.04)
		drawVHSLineGap(((-milliTime - glitchBuffer.height / 3) % glitchBuffer.height).toDouble().absoluteValue, glitchBuffer.height * 0.07)

		if (ClientConfig.ENABLE_COLOR_NOISE) {
			// color encoding errors (encoder/transmission glitch)
			RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
			RenderSystem.setShader(GameRenderer::getPositionTexColorShader)
			colorGlitchVertexBuffer.bind()
			colorGlitchVertexBuffer.drawWithShader(Matrix4f().also { it.setIdentity() }, Matrix4f().also { it.setIdentity() }, GameRenderer.getPositionTexColorShader()!!)
		}

		// upload final result to main frame buffer
		minecraft.mainRenderTarget.bindWrite(true)

		RenderSystem.setShader(GameRenderer::getPositionTexShader)

		RenderSystem.blendFunc(GlStateManager.SourceFactor.ONE, GlStateManager.DestFactor.ZERO)
		RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
		RenderSystem.setShaderTexture(0, glitchBuffer.colorTextureId)
		draw(0.0, 0.0)

		RenderSystem.setProjectionMatrix(projection)
		RenderSystem.getModelViewStack().popPose()
		RenderSystem.applyModelViewMatrix()
	}
}
